/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2ora;

import ch.ehi.basics.logging.EhiLogger;
import ch.interlis.ili2c.metamodel.TransferDescription;
import ch.interlis.ili2c.ModelScan;
import ch.interlis.ili2c.config.Configuration;
import ch.ehi.sqlgen.repository.DbSchema;
import ch.ehi.sqlgen.generator.*;
import ch.ehi.sqlgen.generator.Generator;
import ch.ehi.sqlgen.generator.TransferToSql;
import ch.interlis.iom.swig.iom_javaConstants;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import java.sql.Connection;
import java.sql.DriverManager;

/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 07.02.2005 $
 */
public class Main {
	private static String version=null;
	static public void main(String args[]){
		// arguments on import
		// - modeldir
		// - model(s)/XTF
		// - xtffile
		// - dbhost/name/port/usr/passwd
		String modeldir=null;
		String models=null;
		String createscript=null;
		String dropscript=null;
		String dbhost=null;
		String dbname=null;
		String dbport="1521";
		String dbusr=null;
		String dbpwd=null;
		String xtffile=null;
		String mappingConfig=null;
		if(args.length==0){
			ch.ehi.ili2ora.gui.Main.main(dbhost,dbport,dbname,dbusr,dbpwd,modeldir,models,createscript,dropscript,xtffile,mappingConfig);
			return;
		}
		int argi=0;
		boolean doGui=false;
		for(;argi<args.length;argi++){
			String arg=args[argi];
			if(arg.equals("-modeldir")){
				argi++;
				modeldir=args[argi];
			}else if(arg.equals("-models")){
				argi++;
				models=args[argi];
			}else if(arg.equals("-gui")){
				doGui=true;
			}else if(arg.equals("-trace")){
				EhiLogger.getInstance().setTraceFiler(false); 
			}else if(arg.equals("-dropscript")){
				argi++;
				dropscript=args[argi];
			}else if(arg.equals("-createscript")){
				argi++;
				createscript=args[argi];
			}else if(arg.equals("-dbhost")){
				argi++;
				dbhost=args[argi];
			}else if(arg.equals("-dbname")){
				argi++;
				dbname=args[argi];
			}else if(arg.equals("-dbport")){
				argi++;
				dbport=args[argi];
			}else if(arg.equals("-dbusr")){
				argi++;
				dbusr=args[argi];
			}else if(arg.equals("-dbpwd")){
				argi++;
				dbpwd=args[argi];
			}else if(arg.equals("-mapconfig")){
				argi++;
				mappingConfig=args[argi];
			}else if(arg.equals("-version")){
				printVersion();
				return;
			}else if(arg.equals("-help")){
					printVersion ();
					System.err.println();
					printDescription ();
					System.err.println();
					printUsage ();
					System.err.println();
					System.err.println("OPTIONS");
					System.err.println();
					System.err.println("-dbhost hostname      Name of host with oracle service.");
					System.err.println("-dbport port          Port to access oracle service (default: 1521).");
					System.err.println("-dbname sid           Oracle instance to use.");
					System.err.println("-dbusr  username      User name to access oracle.");
					System.err.println("-dbpwd  password      Password of user used to access oracle.");
					System.err.println("-modeldir  path       Path(s) of directories containing ili-files.");
					System.err.println("-models modelname     Name(s) of ili-models to generate an oracle schema for.");
					System.err.println("-createscript filename  Generate a sql script that creates the oracle schema.");
					System.err.println("-dropscript filename  Generate a sql script that drops the generated oracle schema.");
					System.err.println("-mapconfig filename   Name of config file, that controls the schema mapping.");
					System.err.println("-gui                  start GUI.");
					System.err.println("-trace                enable trace messages.");
					System.err.println("-help                 Display this help text.");
					System.err.println("-version              Display the version of ili2ora.");
					System.err.println();
					return;
				
			}else if(arg.startsWith("-")){
				EhiLogger.logAdaption(arg+": unknown option; ignored");
			}else{
				break;
			}
		}
		if(argi<args.length){
			xtffile=args[argi];
		}
		if(doGui){
			ch.ehi.ili2ora.gui.Main.main(dbhost,dbport,dbname,dbusr,dbpwd,modeldir,models,createscript,dropscript,xtffile,mappingConfig);
		}else{
			runImport(dbhost,dbport,dbname,dbusr,dbpwd,modeldir,models,createscript,dropscript,xtffile,mappingConfig);
		}
		
	}
	public static void runImport(
		String dbhost,
		String dbport,
		String dbname,
		String dbusr,
		String dbpwd,
		String modeldir,
		String models,
		String createscript,
		String dropscript,
		String xtffile,
		String mappingConfig
		) {
		ArrayList modelv=new ArrayList();
		if(models==null){
			EhiLogger.logError("no models given");
			return;
		}
		String modelnames[]=models.split(";");
		for(int modeli=0;modeli<modelnames.length;modeli++){
			modelv.add(modelnames[modeli]);				
		}
		if(modelv.size()==0){
			EhiLogger.logError("no models given");
			return;
		}
		if(modeldir==null){
			EhiLogger.logError("no modeldir given");
			return;
		}
		if(dbhost==null){
			EhiLogger.logError("no dbhost given");
			return;
		}
		if(dbname==null){
			EhiLogger.logError("no dbname given");
			return;
		}
		if(dbusr==null){
			EhiLogger.logError("no dbusr given");
			return;
		}
		if(dbpwd==null){
			EhiLogger.logError("no dbpwd given");
			return;
		}
		if(xtffile==null){
			EhiLogger.logError("no xtf-file given");
			return;
		}
		
		// compile required ili files
		EhiLogger.logState("compile models...");
		TransferDescription td=compileIliFiles(modeldir,modelv,null);
		if(td==null){
			EhiLogger.logError("compile of ili-files failed");
			return;
		}
		// open db connection
		Connection conn=null;
		String url = "jdbc:oracle:thin:@" + dbhost + ":" + dbport + ":" + dbname;
		try{
		  DriverManager.registerDriver(new oracle.jdbc.OracleDriver());
		  conn = DriverManager.getConnection(url, dbusr, dbpwd);
		}catch(Exception ex){
			EhiLogger.logError(ex);
			return;
		}

		initIom();
		
		// read mapping file
		Mapping mapping=new Mapping();
		if(mappingConfig!=null){
			mapping.readConfig(mappingConfig);
		}
		// create table structure
		EhiLogger.logState("create table structure...");
		try{
			TransferFromIli trsfFromIli=new TransferFromIli();
			DbSchema schema=trsfFromIli.doit(td,modelv,mapping);
			if(schema==null){
				return;
			}
			GeneratorOracle gen=new GeneratorOracle();
			TransferToSql trsf=new TransferToSql();
			SqlConfig config=new SqlConfig();
			config.setJdbcConnection(conn);
			trsf.visitSchema(config,schema,gen);
			// create script requested by user?
			if(createscript!=null){
				writeScript(createscript,gen.iteratorCreateLines());
			}
			// drop script requested by user?
			if(dropscript!=null){
				writeScript(dropscript,gen.iteratorDropLines());
			}
			// update mapping table
			mapping.updateMappingTable(conn);
			// update inheritance table
			trsfFromIli.updateInheritanceTable(conn);
		}catch(java.io.IOException ex){
			EhiLogger.logError(ex);
			return;
		}
		
		// process xtf files
		EhiLogger.logState("process data file...");
		transferFromXtf(conn,xtffile,mapping,td,dbusr);
		try{
			conn.close();
			EhiLogger.logState("...import done");
		}catch(java.sql.SQLException ex){
			EhiLogger.logError(ex);
		}
	}
	public static void runExport(
		String dbhost,
		String dbport,
		String dbname,
		String dbusr,
		String dbpwd,
		String modeldir,
		String models,
		String xtffile
		) {
		ArrayList modelv=new ArrayList();
		if(models==null){
			EhiLogger.logError("no models given");
			return;
		}
		String modelnames[]=models.split(";");
		for(int modeli=0;modeli<modelnames.length;modeli++){
			modelv.add(modelnames[modeli]);				
		}
		if(modelv.size()==0){
			EhiLogger.logError("no models given");
			return;
		}
		if(modeldir==null){
			EhiLogger.logError("no modeldir given");
			return;
		}
		if(dbhost==null){
			EhiLogger.logError("no dbhost given");
			return;
		}
		if(dbname==null){
			EhiLogger.logError("no dbname given");
			return;
		}
		if(dbusr==null){
			EhiLogger.logError("no dbusr given");
			return;
		}
		if(dbpwd==null){
			EhiLogger.logError("no dbpwd given");
			return;
		}
		if(xtffile==null){
			EhiLogger.logError("no xtf-file given");
			return;
		}
		
		// compile required ili files
		EhiLogger.logState("compile models...");
		String ilxfile=null;
		try{
			ilxfile=java.io.File.createTempFile("ili2ora","ilx").getAbsolutePath();
		}catch(java.io.IOException ex){
			EhiLogger.logError(ex);
			return;
		}
		TransferDescription td=compileIliFiles(modeldir,modelv,ilxfile);
		if(td==null){
			EhiLogger.logError("compile of ili-files failed");
			return;
		}
		// open db connection
		Connection conn=null;
		String url = "jdbc:oracle:thin:@" + dbhost + ":" + dbport + ":" + dbname;
		try{
		  DriverManager.registerDriver(new oracle.jdbc.OracleDriver());
		  conn = DriverManager.getConnection(url, dbusr, dbpwd);
		}catch(Exception ex){
			EhiLogger.logError(ex);
			return;
		}

		initIom();
		
		// read mapping from db
		Mapping mapping=new Mapping();
		mapping.readMappingTable(conn);
		
		// process xtf files
		EhiLogger.logState("process data...");
		transferToXtf(conn,xtffile,mapping,td,ilxfile);
		try{
			conn.close();
			EhiLogger.logState("...export done");
		}catch(java.sql.SQLException ex){
			EhiLogger.logError(ex);
		}
	}
	static private void writeScript(String filename,Iterator linei)
	throws java.io.IOException
	{
		java.io.PrintWriter out=new java.io.PrintWriter(new java.io.BufferedOutputStream(new java.io.FileOutputStream(filename)));
		while(linei.hasNext()){
		  GeneratorOracle.AbstractStmt stmt=(GeneratorOracle.AbstractStmt)linei.next();
		  String line=stmt.getLine();
		  if(stmt instanceof GeneratorOracle.Stmt){
			out.println(line+";");
		  }else{
			out.println("-- "+line);
		  }
		}
		out.close();				
	}

	/** transfer data from xml file to database
	*/
	static private void transferFromXtf(Connection conn,String xtffile,Mapping ili2sqlName,TransferDescription td,String dbusr){	
		ch.interlis.iom.IomFile iomFile=ch.interlis.iom.IomFile.open(xtffile,0,null);
		TransferFromXtf trsfr=new TransferFromXtf(ili2sqlName,td,conn,dbusr);
		trsfr.doit(iomFile);
	}
	/** transfer data from database to xml file
	*/
	static private void transferToXtf(Connection conn,String xtffile,Mapping ili2sqlName,TransferDescription td,String ilxfile){	
		ch.interlis.iom.IomFile iomFile=ch.interlis.iom.IomFile.open(xtffile,iom_javaConstants.IOM_CREATE | iom_javaConstants.IOM_DONTREAD,null);
		ch.interlis.iom.IomFile iomIlx=ch.interlis.iom.IomFile.open(ilxfile,0,null);
		iomFile.setmodel(iomIlx.iteratorbasket().nextbasket());
		TransferToXtf trsfr=new TransferToXtf(ili2sqlName,td,conn);
		trsfr.doit(iomFile);
		iomFile.save();
	}
	static private void initIom(){
			// setup iom
			ch.interlis.iom.Iom.setLibraryHint("ili2ora.jar");
			ch.interlis.iom.Iom.init();
	}
	static private void endIom(){
			ch.interlis.iom.Iom.end();
	}
	
	/** returns null if compiler fails.
	 * @param ilxFile may be null.
	 */
	static public TransferDescription compileIliFiles(String modeldir,ArrayList models,String ilxFile){
		Configuration config=ModelScan.getConfig(modeldir,models);
		if(config==null){
			return null;
		}
		Iterator filei=config.iteratorFileEntry();
		while(filei.hasNext()){
			ch.interlis.ili2c.config.FileEntry file=(ch.interlis.ili2c.config.FileEntry)filei.next();
			EhiLogger.logState(file.getFilename());
		}
		if(ilxFile!=null){
			config.setOutputKind(ch.interlis.ili2c.config.GenerateOutputKind.IOM);
			config.setOutputFile(ilxFile);
		}else{
			config.setOutputKind(ch.interlis.ili2c.config.GenerateOutputKind.NOOUTPUT);
		}
		TransferDescription ret=ch.interlis.ili2c.Main.runCompiler(config);
		return ret;
	}
	static private HashSet getModelNames(ArrayList modelv){
		HashSet ret=new HashSet();
		Iterator modeli=modelv.iterator();
		while(modeli.hasNext()){
			String topic=(String)modeli.next(); // modelname or qualified topicname
			int endModelName=topic.indexOf('.');
			if(endModelName<=0){
				// just a modelname
				ret.add(topic);
			}else{
				// qualified topicname; get model name
				String model=topic.substring(0,endModelName);
				ret.add(model);
			}
		}
		return ret;
	}
	protected static void printVersion ()
	{
	  System.err.println("INTERLIS 2-loader for oracle, Version "+getVersion());
	  System.err.println("  Developed by Eisenhut Informatik AG, CH-3303 Jegenstorf");
	  System.err.println("  See http://www.interlis.ch for information about INTERLIS");
	  System.err.println("  Parts of this program have been generated by ANTLR; see http://www.antlr.org");
	  System.err.println("  This product includes software developed by the");
	  System.err.println("  Apache Software Foundation (http://www.apache.org/).");
	}


	protected static void printDescription ()
	{
	  System.err.println("DESCRIPTION");
	  System.err.println("  Translates INTERLIS 2 data model definitions to an oracle schema.");
	  System.err.println("  Loads INTERLIS 2 data into an oracle database.");
	  System.err.println("  Unloads INTERLIS 2 data from an oracle database.");
	}


	protected static void printUsage()
	{
	  System.err.println ("USAGE");
	  System.err.println("  java -jar ili2ora.jar [Options] file1.xtf ...");
	}
	public static String getVersion() {
		  if(version==null){
//		java.util.ResourceBundle resVersion = java.util.ResourceBundle.getBundle(ch.ehi.basics.i18n.ResourceBundle.class2packagePath(Main.class)+"/Version");
		java.util.ResourceBundle resVersion = java.util.ResourceBundle.getBundle("ch/ehi/ili2ora/Version");
			// Major version numbers identify significant functional changes.
			// Minor version numbers identify smaller extensions to the functionality.
			// Micro versions are even finer grained versions.
			StringBuffer ret=new StringBuffer(20);
		ret.append(resVersion.getString("versionMajor"));
			ret.append('.');
		ret.append(resVersion.getString("versionMinor"));
			ret.append('.');
		ret.append(resVersion.getString("versionMicro"));
			ret.append('-');
		ret.append(resVersion.getString("versionDate"));
			version=ret.toString();
		  }
		  return version;
	}
}
