/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2ora;

import java.util.ArrayList;
import java.util.Vector;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.sql.PreparedStatement;
import java.sql.Connection;
import java.sql.Types;
import java.sql.ResultSet;

import ch.ehi.basics.logging.EhiLogger;

import ch.interlis.ili2c.metamodel.*;

import ch.interlis.iom.swig.iom_javaConstants;
import ch.interlis.iom.*;
import oracle.spatial.geometry.JGeometry;

/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 06.05.2005 $
 */
public class TransferToXtf {
	private Mapping ili2sqlName=null;
	private TransferDescription td=null;
	private Connection conn=null;
	public TransferToXtf(Mapping ili2sqlName1,TransferDescription td1,Connection conn1){
		ili2sqlName=ili2sqlName1;
		td=td1;
		conn=conn1;
	}
	public void doit(IomFile iomFile)
	{
		// for all MODELs
		Iterator modeli = td.iterator ();
		while (modeli.hasNext ())
		{
		  Object mObj = modeli.next ();
		  if ((mObj instanceof Model) && !(suppressModel((Model)mObj)))
		  {
			Model model=(Model)mObj;
			// for all TOPICs
			Iterator topici=model.iterator();
			while(topici.hasNext()){
				Object tObj=topici.next();
				if (tObj instanceof Topic && !(suppressTopic((Topic)tObj))){
					Topic topic=(Topic)tObj;
					IomBasket iomBasket=null;
					// for all Viewables
					Iterator iter = topic.getViewables().iterator();
					while (iter.hasNext())
					{
					  Object obj = iter.next();
					  if ((obj instanceof Viewable) && !suppressViewable ((Viewable)obj))
					  {
					  	Viewable aclass=(Viewable)obj;
					  	// get sql name
					  	String sqlName=getSqlTableName(aclass);
					  	// if table exists?
					  	if(sqlTableExists(sqlName)){
							// dump it
							EhiLogger.logState(aclass.getScopedName(null)+"...");
							if(iomBasket==null){
								iomBasket=IomBasket.newbasket(iomFile);
								iomBasket.setbaskettag(topic.getScopedName(null));
								iomBasket.setbasketoid(topic.getScopedName(null));
							}
							dumpObject(iomBasket,aclass);
					  	}else{
					  		// skip it
							EhiLogger.traceUnusualState(aclass.getScopedName(null)+"...skipped; no table "+sqlName+" in db");
					  	}
					  }
					}
				}
			}
		  }
		}
	}
	/** dumps all objects of a given class.
	 */
	private void dumpObject(IomBasket iomBasket,Viewable aclass)
	{
		String stmt=createQueryStmt(aclass);
		EhiLogger.traceBackendCmd(stmt);
		java.sql.Statement dbstmt = null;
		try{
			
			dbstmt = conn.createStatement();
			java.sql.ResultSet rs=dbstmt.executeQuery(stmt);
			while(rs.next()){
				String tid=rs.getString("T_Id");
				IomObject iomObj=IomObject.newobject(iomBasket,aclass.getScopedName(null),tid);

				Iterator iter = aclass.getAttributesAndRoles2();
				while (iter.hasNext()) {
				   Viewable.TransferElement obj = (Viewable.TransferElement)iter.next();
				   if (obj.obj instanceof AttributeDef) {
					   AttributeDef attr = (AttributeDef) obj.obj;
					   if(attr.getExtending()==null){
						String attrName=attr.getName();
						Type type = attr.getDomain();
						if( (type instanceof TypeAlias) 
							&& ((TypeAlias) type).getAliasing() == td.INTERLIS.BOOLEAN) {
								int value=rs.getInt(attrName);
								if(!rs.wasNull()){
									if(value==1){
										iomObj.setattrvalue(attrName,"true");
									}else{
										iomObj.setattrvalue(attrName,"false");
									}
								}
						}else{
							type = attr.getDomainResolvingAliases();
							if (type instanceof CompositionType){
								// TODO handle structs
							}else if (type instanceof PolylineType){
								Object geomobj=rs.getObject(attrName);
								if(!rs.wasNull()){
									JGeometry geometry = JGeometry.load((oracle.sql.STRUCT)geomobj);
									if(geometry.getType()!=JGeometry.GTYPE_MULTICURVE){
										EhiLogger.logError("Object "+tid+": unexpected GTYPE ("+gtype2str(geometry.getType())+") in attribute "+attrName);
									}else{
										int dim=geometry.getDimensions();
										if(dim!=((CoordType)((PolylineType)type).getControlPointDomain().getType()).getDimensions().length){
											EhiLogger.logError("Object "+tid+": unexpected dimension ("+Integer.toString(dim)+") in attribute "+attrName);
										}else{
											boolean is3D=(dim==3);
											int elev[]=geometry.getElemInfo();
											double ordv[]=geometry.getOrdinatesArray();
											final int SDO_STARTING_OFFSET=0;
											final int SDO_ETYPE=1;
											final int SDO_INTERPRETATION=2;
											final int NEXT_TRIPLET=3;
											final int COMPOUND_LINESTRING=4;
											final int LINESTRING=2;
											final int STRAIGHT=1;
											final int ARC=2;
											int elei=0;
											if(elev[elei+SDO_ETYPE]!=COMPOUND_LINESTRING){
												EhiLogger.logError("Object "+tid+": unexpected SDO_ETYPE ("+Integer.toString(elev[elei+SDO_ETYPE])+") in attribute "+attrName);
											}else{
												elei+=NEXT_TRIPLET;
												IomObject polylineValue=iomObj.addattrobj(attrName,"POLYLINE");
												// unclipped polyline, add one sequence
												IomObject sequence=polylineValue.addattrobj("sequence","SEGMENTS");
												while(elei<elev.length){
													if(elev[elei+SDO_ETYPE]!=LINESTRING){
														EhiLogger.logError("Object "+tid+": unexpected SDO_ETYPE ("+Integer.toString(elev[elei+SDO_ETYPE])+") in attribute "+attrName);
													}else{
														if(elev[elei+SDO_INTERPRETATION]==STRAIGHT){
															int start=elev[elei+SDO_STARTING_OFFSET]-1;
															int end;
															if(elei+NEXT_TRIPLET>=elev.length){
																end=ordv.length;
															}else{
																end=elev[elei+NEXT_TRIPLET+SDO_STARTING_OFFSET]-1;
															}
															for(int i=start;i<end;){
																// add control point
																IomObject coordValue=sequence.addattrobj("segment","COORD");
																coordValue.setattrvalue("C1",Double.toString(ordv[i]));
																coordValue.setattrvalue("C2",Double.toString(ordv[i+1]));
																if(is3D){
																	coordValue.setattrvalue("C3",Double.toString(ordv[i+2]));
																	i+=3;
																}else{
																	i+=2;
																}
															}
														}else if(elev[elei+SDO_INTERPRETATION]==ARC){
															int start=elev[elei+SDO_STARTING_OFFSET]-1;
															int end;
															if(elei+NEXT_TRIPLET>=elev.length){
																end=ordv.length;
															}else{
																end=elev[elei+NEXT_TRIPLET+SDO_STARTING_OFFSET]-1;
															}
															for(int i=start;i<end;){
																// add control point
																IomObject coordValue=sequence.addattrobj("segment","ARC");
																coordValue.setattrvalue("A1",Double.toString(ordv[i]));
																coordValue.setattrvalue("A2",Double.toString(ordv[i+1]));
																if(is3D){
																	// no A3 in XTF!
																	i+=3;
																}else{
																	i+=2;
																}
																coordValue.setattrvalue("C1",Double.toString(ordv[i]));
																coordValue.setattrvalue("C2",Double.toString(ordv[i+1]));
																if(is3D){
																	coordValue.setattrvalue("C3",Double.toString(ordv[i+2]));
																	i+=3;
																}else{
																	i+=2;
																}
															}
														}else{
															EhiLogger.logError("Object "+tid+": unexpected SDO_INTERPRETATION ("+Integer.toString(elev[elei+SDO_INTERPRETATION])+") in attribute "+attrName);
														}
													}
													elei+=NEXT_TRIPLET;
												}
											}
										}
									
									}
								}
							 }else if(type instanceof SurfaceOrAreaType){
								Object geomobj=rs.getObject(attrName);
								if(!rs.wasNull()){
									JGeometry geometry = JGeometry.load((oracle.sql.STRUCT)geomobj);
									if(geometry.getType()!=JGeometry.GTYPE_POLYGON){
										EhiLogger.logError("Object "+tid+": unexpected GTYPE ("+gtype2str(geometry.getType())+") in attribute "+attrName);
									}else{
										int dim=geometry.getDimensions();
										if(dim!=((CoordType)((SurfaceOrAreaType)type).getControlPointDomain().getType()).getDimensions().length){
											EhiLogger.logError("Object "+tid+": unexpected dimension ("+Integer.toString(dim)+") in attribute "+attrName);
										}else{
											boolean is3D=(dim==3);
											int elev[]=geometry.getElemInfo();
											double ordv[]=geometry.getOrdinatesArray();
											final int SDO_STARTING_OFFSET=0;
											final int SDO_ETYPE=1;
											final int SDO_INTERPRETATION=2;
											final int NEXT_TRIPLET=3;
											final int EXTERIOR_POLYGON_RING=1005; // (must be specified in counterclockwise order)
											final int INTERIOR_POLYGON_RING=2005; // (must be specified in clockwise order)
											final int LINESTRING=2;
											final int STRAIGHT=1;
											final int ARC=2;
											int elei=0;
											IomObject multisurface=iomObj.addattrobj(attrName,"MULTISURFACE");
											IomObject surface=multisurface.addattrobj("surface","SURFACE");
											while(elei<elev.length){
												if(elev[elei+SDO_ETYPE]!=EXTERIOR_POLYGON_RING && elev[elei+SDO_ETYPE]!=INTERIOR_POLYGON_RING){
													EhiLogger.logError("Object "+tid+": unexpected SDO_ETYPE ("+Integer.toString(elev[elei+SDO_ETYPE])+") in attribute "+attrName);
												}else{
													int nTriplet=elev[elei+SDO_INTERPRETATION];
													elei+=NEXT_TRIPLET;
												
													IomObject boundary=surface.addattrobj("boundary","BOUNDARY");
													IomObject polylineValue=boundary.addattrobj("polyline","POLYLINE");
													IomObject sequence=polylineValue.addattrobj("sequence","SEGMENTS");
													for(int iTriplet=0;iTriplet<nTriplet;iTriplet++){
														if(elev[elei+SDO_ETYPE]!=LINESTRING){
															EhiLogger.logError("Object "+tid+": unexpected SDO_ETYPE ("+Integer.toString(elev[elei+SDO_ETYPE])+") in attribute "+attrName);
														}else{
															if(elev[elei+SDO_INTERPRETATION]==STRAIGHT){
																int start=elev[elei+SDO_STARTING_OFFSET]-1;
																int end;
																if(elei+NEXT_TRIPLET>=elev.length){
																	end=ordv.length;
																}else{
																	end=elev[elei+NEXT_TRIPLET+SDO_STARTING_OFFSET]-1;
																}
																for(int i=start;i<end;){
																	// add control point
																	IomObject coordValue=sequence.addattrobj("segment","COORD");
																	coordValue.setattrvalue("C1",Double.toString(ordv[i]));
																	coordValue.setattrvalue("C2",Double.toString(ordv[i+1]));
																	if(is3D){
																		coordValue.setattrvalue("C3",Double.toString(ordv[i+2]));
																		i+=3;
																	}else{
																		i+=2;
																	}
																}
															}else if(elev[elei+SDO_INTERPRETATION]==ARC){
																int start=elev[elei+SDO_STARTING_OFFSET]-1;
																int end;
																if(elei+NEXT_TRIPLET>=elev.length){
																	end=ordv.length;
																}else{
																	end=elev[elei+NEXT_TRIPLET+SDO_STARTING_OFFSET]-1;
																}
																for(int i=start;i<end;){
																	// add control point
																	IomObject coordValue=sequence.addattrobj("segment","ARC");
																	coordValue.setattrvalue("A1",Double.toString(ordv[i]));
																	coordValue.setattrvalue("A2",Double.toString(ordv[i+1]));
																	if(is3D){
																		// no A3 in XTF!
																		i+=3;
																	}else{
																		i+=2;
																	}
																	coordValue.setattrvalue("C1",Double.toString(ordv[i]));
																	coordValue.setattrvalue("C2",Double.toString(ordv[i+1]));
																	if(is3D){
																		coordValue.setattrvalue("C3",Double.toString(ordv[i+2]));
																		i+=3;
																	}else{
																		i+=2;
																	}
																}
															}else{
																EhiLogger.logError("Object "+tid+": unexpected SDO_INTERPRETATION ("+Integer.toString(elev[elei+SDO_INTERPRETATION])+") in attribute "+attrName);
															}
														}
														elei+=NEXT_TRIPLET;
													}
												}
											}
										}
									
									}
								}
							 }else if(type instanceof CoordType){
								Object geomobj=rs.getObject(attrName);
								if(!rs.wasNull()){
									JGeometry geometry = JGeometry.load((oracle.sql.STRUCT)geomobj);
									if(geometry.getType()!=JGeometry.GTYPE_POINT){
										EhiLogger.logError("Object "+tid+": unexpected GTYPE ("+gtype2str(geometry.getType())+") in attribute "+attrName);
									}else{
										int dim=geometry.getDimensions();
										if(dim!=((CoordType)type).getDimensions().length){
											EhiLogger.logError("Object "+tid+": unexpected dimension ("+Integer.toString(dim)+") in attribute "+attrName);
										}else{
											double[] valuev=geometry.getFirstPoint();
											IomObject coord=iomObj.addattrobj(attrName,"COORD");
											coord.setattrvalue("C1",Double.toString(valuev[0]));
											coord.setattrvalue("C2",Double.toString(valuev[1]));
											if(dim==3){
												coord.setattrvalue("C3",Double.toString(valuev[2]));
											}
										}
									
									}
								}
							}else{
								String value=rs.getString(attrName);
								if(!rs.wasNull()){
									iomObj.setattrvalue(attrName,value);
								}
							}
						   }
						}
				   }
				   if(obj.obj instanceof RoleDef){
					   RoleDef role = (RoleDef) obj.obj;
					   if(role.getExtending()==null){
						String roleName=role.getName();
						// a role of an embedded association?
						if(obj.embedded){
							AssociationDef roleOwner = (AssociationDef) role.getContainer();
							if(roleOwner.getDerivedFrom()==null){
								 // TODO if(orderPos!=0){
								IomObject ref=iomObj.addattrobj(roleName,roleOwner.getScopedName(null));
								ref.setobjectrefoid(rs.getString(roleName));
							}
						 }else{
							 // TODO if(orderPos!=0){
							IomObject ref=iomObj.addattrobj(roleName,"REF");
							ref.setobjectrefoid(rs.getString(roleName));
						 }
					   }
					}
				}
				
			}
		}catch(java.sql.SQLException ex){		
			EhiLogger.logError("failed to query "+aclass.getScopedName(null),ex);
		}finally{
			if(dbstmt!=null){
				try{
					dbstmt.close();
				}catch(java.sql.SQLException ex){		
					EhiLogger.logError("failed to close query of "+aclass.getScopedName(null),ex);
				}
			}
		}
		
	}
	private String gtype2str(int gtype)
	{
		switch(gtype){
			case JGeometry.GTYPE_COLLECTION:
				return "GTYPE_COLLECTION";
			case JGeometry.GTYPE_CURVE:
				return "GTYPE_CURVE";
			case JGeometry.GTYPE_MULTICURVE:
				return "GTYPE_MULTICURVE";
			case JGeometry.GTYPE_MULTIPOINT:
				return "GTYPE_MULTIPOINT";
			case JGeometry.GTYPE_MULTIPOLYGON:
				return "GTYPE_MULTIPOLYGON";
			case JGeometry.GTYPE_POINT:
				return "GTYPE_POINT";
			case JGeometry.GTYPE_POLYGON:
				return "GTYPE_POINT";
			default:
				return Integer.toString(gtype);
		}
	}
	/** tests if a table with the given name exists
	 */
	private boolean sqlTableExists(String tableName)
	{
		try{
			String catalog=conn.getCatalog();
			java.sql.DatabaseMetaData meta=conn.getMetaData();
			// on oracle getUserName() returns schemaname
			String schema=meta.getUserName();
			ResultSet rs=null;
			try{
				rs=meta.getTables(catalog,schema,tableName.toUpperCase(),null);
				if(rs.next()){
					// table exists
					return true;
				}
			}finally{
				if(rs!=null)rs.close();
			}
		}catch(java.sql.SQLException ex){
			EhiLogger.logError("failed to check if table "+tableName+" exists",ex);
		}
		return false;
	}
	protected boolean suppressModel (Model model)
	{
	  if (model == null)
		return true;


	  if (model == td.INTERLIS)
		return true;


	  if ((model instanceof TypeModel) || (model instanceof PredefinedModel))
		return true;


	  return false;
	}
	protected boolean suppressTopic (Topic topic)
	{
	  if (topic == null)
		return true;


	  if (topic.isAbstract ())
		return true;


	  return false;
	}
	protected boolean suppressViewable (Viewable v)
	{
	  if (v == null)
		return true;


	  if (v.isAbstract())
		return true;

	  if(v instanceof AssociationDef){
		  AssociationDef assoc=(AssociationDef)v;
		  if(assoc.isLightweight()){
			  return true;
		  }
		  if(assoc.getDerivedFrom()!=null){
			  return true;
		  }
	  }

	  Topic topic;
	  if ((v instanceof View) && ((topic=(Topic)v.getContainer (Topic.class)) != null)
		  && !topic.isViewTopic())
		return true;


	  /* STRUCTUREs do not need to be printed with their INTERLIS container,
		 but where they are used. */
	  if ((v instanceof Table) && !((Table)v).isIdentifiable())
		return true;


	  return false;
	}
	/** maps ili class name to a sql table name.
	 */
	private String getSqlTableName(Viewable def){
		String sqlname=ili2sqlName.mapIliClassDef(def);
		return sqlname;
	}
	/** creates sql query statement for a class.
	 */
	private String createQueryStmt(Viewable aclass){
		StringBuffer ret = new StringBuffer();
		ret.append("SELECT r0.T_Id, r0.T_Type");
		String sep=",";
		Iterator iter = aclass.getAttributesAndRoles2();
		while (iter.hasNext()) {
		   Viewable.TransferElement obj = (Viewable.TransferElement)iter.next();
		   if (obj.obj instanceof AttributeDef) {
			   AttributeDef attr = (AttributeDef) obj.obj;
			   if(attr.getExtending()==null){
				Type type = attr.getDomainResolvingAliases();
				 String attrName=attr.getName();
				if (type instanceof CompositionType){
				}else if (type instanceof PolylineType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
				 }else if(type instanceof SurfaceType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
				 }else if(type instanceof AreaType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
				 }else if(type instanceof CoordType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
				}else{
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
				}
			   }
		   }
		   if(obj.obj instanceof RoleDef){
			   RoleDef role = (RoleDef) obj.obj;
			   if(role.getExtending()==null){
				String roleName=role.getName();
				// a role of an embedded association?
				if(obj.embedded){
					AssociationDef roleOwner = (AssociationDef) role.getContainer();
					if(roleOwner.getDerivedFrom()==null){
						 // TODO if(orderPos!=0){
						 ret.append(sep);
						 sep=",";
						 ret.append(roleName);
					}
				 }else{
					 // TODO if(orderPos!=0){
					 ret.append(sep);
					 sep=",";
					 ret.append(roleName);
				 }
			   }
			}
		}
		// stdcols
		ret.append(sep);
		sep=",";
		ret.append("r0.T_LastChange");
		ret.append(sep);
		sep=",";
		ret.append("r0.T_CreateDate");
		ret.append(sep);
		sep=",";
		ret.append("r0.T_User");

		ret.append(" FROM ");
		ArrayList tablev=new ArrayList(10);
		tablev.add(aclass);
		Viewable base=(Viewable)aclass.getExtending();
		while(base!=null){
			tablev.add(base);		
			base=(Viewable)base.getExtending();
		}
		sep="";
		int tablec=tablev.size();
		for(int i=0;i<tablec;i++){
			ret.append(sep);
			ret.append(getSqlTableName((Viewable)tablev.get(i)));
			ret.append(" r"+Integer.toString(tablec-1-i));
			if(i>0){
				ret.append(" ON r"+Integer.toString(tablec-i)+".t_id=r"+Integer.toString(tablec-1-i)+".t_id");
			}
			sep=" INNER JOIN ";
		}
		ret.append(" WHERE r0.T_Type='"+getSqlTableName(aclass)+"'");
		return ret.toString();
	}

}
